#include "forensic.h"

using namespace forensic;

/**
 * Creates a Config object with a default configuration.
 *
 * The default configuration can be obtained from the compiled code by
 * asking Exgrep to dump its configuration to file.
 */
Config::Config()
{
  maxLength = 0;
  
  setVerbose(0);
  
  // Setting the default configuration
  // GLOBAL settings
  setStartPos(0);
  setEndPos(0);
  setSectorSize(0);
  setActiveGroups("");
  
  // GROUPS settings
  defineGroup("IMAGES", "jpg|gif|tif|png");
  defineGroup("DOCUMENTS", "doc");
  defineGroup("WEB_DOCUMENTS", "pdf|eps");
  defineGroup("IE_CACHE", "ie_hist|url");
  /* These give too many false positives
  defineGroup("MOVIES", "mov|mpg|mpeg|avi|wmv");
  */
  // RE's
  CutDefinition* cut = new CutDefinition();
  cut->setExtension("jpg");
  cut->addStartRE("/\\xff\\xd8....(JFIF|Exif)/");
  cut->setMaxLength(500000);
  cut->setOutputDirectory("images/jpeg");
  cut->addComment("JPEG picture file type");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("gif");
  cut->addStartRE("/GIF8[79]a/");
  cut->setMaxLength(50000);
  cut->setOutputDirectory("images/gifs");
  cut->addComment("GIF picture file type");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("png");
  cut->addStartRE("/\\x89PNG\\x0d\\x0a\\x1a\\x0a/");
  cut->addEndRE("/\\x45\\x4e\\x44\\xae\\x42\\x60\\x82/");
  cut->setMaxLength(500000);
  cut->setOutputDirectory("images/png");
  cut->addComment("PNG picture file type");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("tif");
  cut->addStartRE("/\\x49\\x49\\x2a\\x00\\x08/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("images/tif");
  cut->addComment("TIF picture file type 1");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("tif");
  cut->addStartRE("/\\x4d\\x4d\\x00\\x2a\\x00/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("images/tif");
  cut->addComment("TIF picture file type 2");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("doc");
  cut->addStartRE("/\\xd0\\xcf\\x11\\xe0/");
  cut->setMaxLength(500000);
  cut->setOutputDirectory("office");
  cut->addComment("MS Word document");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("pdf");
  cut->addStartRE("/%PDF-/");
  cut->addEndRE("/.%%EOF\\x0d/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("pdf");
  cut->addComment("Portable Document Format");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("eps");
  cut->addStartRE("/%!PS-Adobe/");
  cut->addEndRE("/end.%%.trailer/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("eps");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("eps");
  cut->addStartRE("/%!PS-Adobe/");
  cut->addEndRE("/%%EOF./");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("eps");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("ie_hist");
  cut->addStartRE("/Client UrlCache/");
  cut->setMaxLength(300000);
  cut->setOutputDirectory("ie_hist");
  cut->addComment("Internet Explorer URL cache");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("url");
  cut->addStartRE("/URL \\x03\\x00\\x00\\x00/");
  cut->setMaxLength(384);
  cut->setOutputDirectory("ie_hist");
  cut->addComment("Internet Explorer URL cache");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("url");
  cut->addStartRE("/URL \\x02\\x00\\x00\\x00/");
  cut->setMaxLength(256);
  cut->setOutputDirectory("ie_hist");
  cut->addComment("Internet Explorer URL cache");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("mov");
  cut->addStartRE("/....free.....mov/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies");
  cut->addComment("Quicktime MOV format type 1");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("mov");
  cut->addStartRE("/MOVI/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies");
  cut->addComment("Quicktime MOV format type 2");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("mov");
  cut->addStartRE("/....moov/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies");
  cut->addComment("Quicktime MOV format type 3");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("mov");
  cut->addStartRE("/....mdat/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies/");
  cut->addComment("Quicktime MOV format type 4");
  addCutDefinition(*cut);
  delete cut;

  /* These give too many false positives
  cut = new CutDefinition();
  cut->setExtension("mpg");
  cut->addStartRE("/\\x00\\x00\\x01\\xb3/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies");
  cut->addComment("MPEG format type 1");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("mpeg");
  cut->addStartRE("/\\x00\\x00\\x01\\xba/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies");
  cut->addComment("MPEG format type 2");
  addCutDefinition(*cut);
  delete cut;
  */
  
  cut = new CutDefinition();
  cut->setExtension("avi");
  cut->addStartRE("/RIFF....AVI/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies");
  cut->addComment("AVI video format");
  addCutDefinition(*cut);
  delete cut;
  
  cut = new CutDefinition();
  cut->setExtension("wmv");
  cut->addStartRE("/\\x30\\x26\\xb2\\x75\\x8e\\x66/");
  cut->setMaxLength(1000000);
  cut->setOutputDirectory("movies");
  cut->addComment("Windows movie file");
  addCutDefinition(*cut);
  delete cut;
  
  if (isValid() == false)
  {
    cerr << "ERROR!  Default config is not valid" << endl;
    exit(-1);
  }
}

/**
 * Duplicates the supplied Configuration object.
 *
 * \param config The Config object to be duplicated.
 */
Config::Config(const Config& config)
{
  verbose = config.verbose;
  startPos = config.startPos;
  endPos = config.endPos;
  sectorSize = config.sectorSize;
  activeGroups = config.activeGroups;
  groups = config.groups;
  cuts = config.cuts;
  activeCuts = config.activeCuts;
  
  maxLength = config.maxLength;
}

/**
 * Destroys the Config object.
 */
Config::~Config()
{
  if (getVerbose() > 2)
  {
    cout << "Destroying configuration object." << endl;
  }
}

/**
 * Makes this config object equal the config object provided.
 *
 * \param config The object to be duplicated.
 * \returns A pointer to this.
 */
Config& Config::operator=(const Config& config)
{
  if (this == &config)
  {
    return *this;
  }
  
  verbose = config.verbose;
  startPos = config.startPos;
  endPos = config.endPos;
  sectorSize = config.sectorSize;
  activeGroups = config.activeGroups;
  groups = config.groups;
  cuts.clear();
  cuts = config.cuts;
  activeCuts.clear();
  activeCuts = config.activeCuts;
  
  maxLength = config.maxLength;
  
  return *this;
}

/**
 * Clears the config to an 'empty' state.  This means that the configuration
 * - Has no start or end position
 * - Has a sector size of 0
 * - Has no cut definitions, no groups
 * - All cut definitions that it receives in the future will be active by default.
 *
 * Note that the value of the verbose flag is unaffected.
 */
void Config::clear()
{
  // Setting everything to a clean state
  // Verbosity is retained
  if (getVerbose() > 1)
  {
    cout << "Config cleared."  << endl;
  }
  startPos = 0;
  endPos = 0;
  sectorSize = 0;
  activeGroups = "";
  groups.clear();
  cuts.clear();
  activeCuts.clear();
}

/**
 * Sets the verbosity of the program.
 *
 * Default verboseness is 0.  It can be increased up to 3.
 *
 * \param value The new verbose value.
 */
void Config::setVerbose(int value)
{
  if ((value > -1) && (value < 4))
  {
    verbose = value;
    if (getVerbose() > 1)
    {
      cout << "Verboseness set to " << getVerbose() << endl;
    }
  } else
  {
    cerr << "Valid verbose values: 0 1 2 3" << endl;
  }
}

/**
 * Returns the verbose value.
 *
 * \returns The verbose value.
 */
int Config::getVerbose() const
{
  return verbose;
}

/**
 * Sets the starting position.
 *
 * The start position is defined as the
 * first byte to examine in the input stream. (ie, skip all bytes before
 * this offset).  A value of zero indicates that examination should start
 * once the first byte has been received.
 *
 * \param pos The starting offset.
 */
void Config::setStartPos(unsigned long long pos)
{
  if (getVerbose() > 1)
  {
    cout << "Start position set to " << pos << endl;
  }
  startPos = pos;
}

/**
 * Returns the starting position.
 *
 * \returns The starting position.
 */
unsigned long long Config::getStartPos() const
{
  return startPos;
}

/**
 * Sets the ending position.
 *
 * The end position is defined as the last byte to examine in the input stream
 * (ie stop processing once this byte has been reached)
 * A value of zero indicates that all bytes after the start pos should be examined.
 *
 * \param pos The ending offset.
 */
void Config::setEndPos(unsigned long long pos)
{
  if (getVerbose() > 1)
  {
    cout << "Ending position set to " << pos << endl;
  }
  endPos = pos;
}

/**
 * Returns the ending position.
 *
 * \returns The ending position.
 */
unsigned long long Config::getEndPos() const
{
  return endPos;
}

/**
 * Sets the active groups.
 *
 * The active groups string can be a series of groups separated by vertical 
 * pipes '|' or a series of cut definition extensions separated by vertical
 * pipes '|' or a mix of both.  Where a group name is identical to a cut
 * definition extension it will be assumed that the group name was the one intended.
 *
 * If the string is empty (a zero length string) it will be assumed that all
 * cut definitions should be activated.
 *
 * \param active The string detailing the active groups.
 */
void Config::setActiveGroups(const string& active)
{
  if (getVerbose() > 1)
  {
    cout << "Active groups are now: " << active << endl;
  }
  activeGroups = active;
}

/**
 * Returns a pointer to the string containing the active groups.
 *
 * Note, the pointer should not be deleted.
 *
 * \returns A pointer to the string containing the active groups.
 */
const string* Config::getActiveGroups() const
{
  return &activeGroups;
}

/**
 * Sets the sector size for the input stream.
 *
 * This means that only cut definitions that start at a multiple of this sector size
 * will be considered valid.  Setting the sector size to 0 means that all matches
 * should be considered valid.
 *
 * \param size The sector size in bytes.
 */
void Config::setSectorSize(int size)
{
  if (getVerbose() > 1)
  {
    cout << "Sector size is now " << size << endl;
  }
  sectorSize = size;
}

/**
 * Returns the sector size in bytes.
 *
 * \returns The sector size in bytes.
 */
int Config::getSectorSize() const
{
  return sectorSize;
}

/**
 * This method is used to define a new group.
 *
 * A group is a collection of extensions that are related, and are therefore given
 * a single name to refer to the collection.
 *
 * \param name The name given to the group
 * \param value The list of cut definitions separated by vertical pipes '|'
 * \returns True if the group definition was successful.  False otherwise.
 */
bool Config::defineGroup(const string& name, const string& value)
{
  if (value.length() == 0)
  {
    cerr << "No value provided for " << name << endl;
    return false;
  }
  
  if (getVerbose() > 2)
  {
    cout << "Defining group: " << name << " -> " << value << endl;
  }
  groups[stringToUpper(name.c_str())] = value;
  return true;
}

/**
 * Add a cut definition to the configuration.
 *
 * \param startRE A vector containing all the starting RE's.
 * \param endRE A vector containing all the ending RE's.
 * \param max The maximum length for this cut.
 * \param extension The extension for this cut.
 * \param outputDirectory The directory to put the cuts in.
 * \param comments A vector containing all the comments for this cut.
 * \returns True if the parameters supplied are ok.  False otherwise.
 */
bool Config::addCutDefinition(const vector<string>& startRE, const vector<string>& endRE, unsigned int max, const string& extension, const string& outputDirectory, const vector<string>& comments)
{
  bool allOk = true;
  
  CutDefinition cut;
  
  for (unsigned int i = 0; i < startRE.size(); i++)
  {
    if (cut.addStartRE(startRE[i]) == false)
    {
      allOk = false;
    }
  }
  
  for (unsigned int i = 0; i < endRE.size(); i++)
  {
    if (cut.addEndRE(endRE[i]) == false)
    {
      allOk = false;
    }
  }
  
  cut.setExtension(extension);
  cut.setMaxLength(max);
  cut.setOutputDirectory(outputDirectory);
  
  for (unsigned int i = 0; i < comments.size(); i++)
  {
    cut.addComment(comments[i]);
  }
  
  if (allOk)
  {
    if (getVerbose() > 1)
    {
      cout << "Adding definition for " << extension << endl;
    }
    addCutDefinition(cut);
  }
  
  return allOk;
}

/**
 * Returns all the cut definitions.  This includes active and inactive cut definitions.
 *
 * \returns A pointer to a vector containing cut definitions.
 */
const vector<CutDefinition>* Config::getAllCutDefinitions() const
{
  return &cuts;
}

/**
 * Returns the active cut definitions.
 *
 * \returns A pointer to a vector containing active cut definitions.
 */
const vector<CutDefinition>* Config::getActiveCutDefinitions() const
{
  return &activeCuts;
}

/**
 * Returns the maximum length of all the cut definitions.
 *
 * \returns The maximum length.
 */
unsigned int Config::getMaxLength() const
{
  return maxLength;
}

/**
 * Indicates whether the configuration appears to be complete.
 *
 * This must be run before the configuration is used as it prepares internal
 * structures for use.
 *
 * \returns True if the configuration looks valid, false otherwise.
 */
bool Config::isValid()
{
  bool allOk = true;
  
  if (activateGroups() == false)
  {
    cerr << "Groups not valid" << endl;
    allOk = false;
  }
  
  if ((endPos != 0) && (endPos <= startPos))
  {
    cerr << "The end offset must be greater than the starting offset if something is to occur." << endl;
    allOk = false;
  }
  
  if (cuts.size() == 0)
  {
    cerr << "No regular expressions supplied!" << endl;
    allOk = false;
  }
  
  for (unsigned int i = 0; i < cuts.size(); i++)
  {
    if (cuts[i].isValid() == false)
    {
      cerr << "Cut definition for " << cuts[i].getExtension() << " not valid" << endl;
      allOk = false;
    }
  }
  
  if (activeGroups.length() == 0)
  {
    activeCuts = cuts;
  } else
  {
    if (activateGroups() == false)
    {
      cerr << "Unable to activate the desired groups" << endl;
      allOk = false;
    }
  }
  
  if (activeCuts.size() == 0)
  {
    cerr << "No regular expressions are active!" << endl;
    allOk = false;
  }
  
  if (getVerbose() > 1)
  {
    if (allOk)
    {
      cout << "Config ok!" << endl;
    } else
    {
      cout << "Config NOT ok." << endl;
    }
  }
  
  return allOk;
}

/**
 * Dumps the current running configuration to a file.
 *
 * If the configuration is not valid this would probably output an invalid
 * configuration file. (however the config file would be syntactically valid)
 *
 * \param file The name of the file to write the config to.
 */
void Config::dumpConfig(string file)
{
  ofstream outFile(file.c_str());
  
  if (!outFile)
  {
    cerr << "Could not create config dump file: " << file << endl;
    return;
  }
  
  writeConfigCommentHeader(outFile);
  writeConfigCommentGlobal(outFile);
  writeConfigTextGlobal(outFile);
  writeConfigCommentGroup(outFile);
  writeConfigTextGroup(outFile);
  writeConfigCommentCutDefinition(outFile);
  writeConfigTextCutDefinition(outFile);
  
  outFile.close();
  
  cout << "Config file " << file << " created" << endl;
}

/**
 * Adds a cut definition to the cut definition vector.
 *
 * \param cut The cut definition.
 */
void Config::addCutDefinition(const CutDefinition& cut)
{
  if (cut.getMaxLength() > maxLength)
  {
    maxLength = cut.getMaxLength();
  }
  
  for (unsigned int i = 0; i < cuts.size(); i++)
  {
    if (cuts[i] == cut)
    {
      if (verbose)
      {
        cout << "Replacing old definition for " << cut.getExtension() << endl;
      }
      
      cuts[i] = cut;
      return;
    }
  }
  
  cuts.push_back(cut);
}

/**
 * Reads which groups should be active and attempts to activate them.
 *
 * \returns True if all the groups activate ok.  False otherwise.
 */
bool Config::activateGroups()
{
  activeCuts.clear();
  
  if (activeGroups.length() == 0)
  {
    activeCuts = cuts;
    return true;
  }
  
  bool allOk = true;
  int pos = 0;
  int lastPos = 0;
  
  do
  {
    pos = activeGroups.find_first_of("|", lastPos);
    
    string group = stringToUpper(activeGroups.substr(lastPos, pos - lastPos));
    
    if (groups.find(group) != groups.end())
    {
      if (activateCutDefinition(groups[group]) == false)
      {
        allOk = false;
      }
    } else
    {
      // Wasn't a group.  Perhaps is a cut definition.
      if (activateCutDefinition(group) == false)
      {
        allOk = false;
      }
    }
    
    lastPos = pos + 1;
  } while (pos != -1);
  
  return allOk;
}

/**
 * Writes the header for the config file to the supplied ostream.
 *
 * \param out The output stream.
 */
void Config::writeConfigCommentHeader(ostream& out) const
{
  out << 
    "# This is an example config file for exgrep\n"
    "# \n"
    "# Note:\n"
    "# Tags are all case insensitive.\n"
    "#\n"
    "# Limitations:\n"
    "# Lines cannot contain more than 255 characters\n"
    "# Comments are prefixed with #'s and cannot occur after\n"
    "#   statements (ie statements and comments appear on\n"
    "#   separate lines)\n"
    "# Lines cannot be continued.  Ie, placing a \\\n"
    "#   at the end of a line does not append the next line onto\n"
    "#   this line.\n"
    "############################################################\n"
    "\n";
}

/**
 * Writes the Global comments for the config file to the supplied ostream.
 *
 * \param out The output stream.
 */
void Config::writeConfigCommentGlobal(ostream& out) const
{
  out << 
    "############################################################\n"
    "# The global section defines paramters which are not\n"
    "# specific to particular RE's\n"
    "#\n"
    "# START - optional - Default 0\n"
    "# Defines where exgrep will start within the file\n"
    "# A value of 0 starts at the start of the file\n"
    "#\n"
    "# STOP - optional - Default 0\n"
    "# Defines how far it should read into the file this\n"
    "# allows exgrep to stop after it reaches a certain point.\n"
    "# A value of 0 tells exgrep to read until the end of the\n"
    "# file.  Note that this is NOT a length.  Ie if START is 10\n"
    "# and STOP is 11 it will only read one byte.\n"
    "#\n"
    "# ACTIVE - optional - Default blank (all)\n"
    "# Activates groups of RE's.  By default all RE's will be\n"
    "# activated.  If any groups are specified in this element\n"
    "# then all other groups that aren't specified will be\n"
    "# disabled.  It is also possible to activate RE's\n"
    "# individually provided that the RE name is not obscured by\n"
    "# a group name.  For example, ACTIVE = IMAGES|pdf is valid\n"
    "# and will activate the IMAGE group and the pdf RE.\n"
    "# (assuming that pdf is not the name of a group, in which\n"
    "# case it would enable the IMAGE group and the pdf group)\n"
    "#\n"
    "# SECTOR - optional - Default 0\n"
    "# Sets the sector size to search.  Increases the chances of\n"
    "# good matches (less false positives) but decreases the\n"
    "# chance of finding all cuts (more false negatives).\n"
    "# A value of 0 disables sector boundary searching.  Default\n"
    "# sector boundaries are usually one of 512, 1024 or 2048.\n"
    "#\n";
}

/**
 * Writes the Group comments for the config file to the supplied ostream.
 *
 * \param out The output stream.
 */
void Config::writeConfigCommentGroup(ostream& out) const
{
  out << 
    "############################################################\n"
    "# The groups section allows the bundling of several RE's\n"
    "# into a category.  Categories are user defined but must\n"
    "# conform to <name> = <value> where name should only consist\n"
    "# of alphanumeric characters and the underscore '_'.\n"
    "#\n"
    "# For clarity the names of the groups should be made separate\n"
    "# from the names of the RE's.  So while\n"
    "# jpg = jpg|gif|png\n"
    "# is syntactically valid, it obscures the name 'jpg' and\n"
    "# prevents the jpg RE from being enabled directly from the\n"
    "# ACTIVE tag in the GLOBAL section.\n"
    "# \n";
}

/**
 * Writes the Cut definition comments for the config file to the supplied ostream.
 *
 * \param out The output stream.
 */
void Config::writeConfigCommentCutDefinition(ostream& out) const
{
  out <<
    "############################################################\n"
    "# All other sections describe cut definitions.  The tag\n"
    "# within the square brackets identifies the filename extension.\n"
    "# Filename extensions may be reused.  This allows for files\n"
    "# of the same type to be identified with different RE's\n"
    "# (for example, the TIFF file format has several different\n"
    "# types of header depending on the version of the encoder)\n"
    "#\n"
    "# START - mandatory\n"
    "# Defines the starting RE for the cut.\n"
    "#\n"
    "# END - optional\n"
    "# Defines the RE to match to signal the end of the cut.\n"
    "#\n"
    "# LENGTH - mandatory\n"
    "# Defines the maximum length (in bytes) of the cut to\n"
    "# grab if END RE is not found (or if one is not provided).\n"
    "# Cuts are guaranteed to have this as a maximum length.\n"
    "#\n"
    "# DIRECTORY - optional - Default is present directory\n"
    "# Indicates which directory the new cut files will\n"
    "# be placed into.  If no directory is specified it will be\n"
    "# placed in the current directory.\n"
    "#\n"
    "# COMMENT - optional - One line of free flowing text, however\n"
    "# multiple COMMENT lines may be used to write out a more\n"
    "# detailed comment\n"
    "#\n";
}

/**
 * Writes out the global configuration to the supplied ostream.
 *
 * \param out The output stream.
 */
void Config::writeConfigTextGlobal(ostream& out) const
{
  out << "[GLOBAL]" << endl
      << "START=" << startPos << endl
      << "END=" << endPos << endl
      << "ACTIVE=" << activeGroups << endl
      << "SECTOR=" << sectorSize << endl
      << endl;
}

/**
 * Writes out the group configuration to the supplied ostream.
 *
 * \param out The output stream.
 */
void Config::writeConfigTextGroup(ostream& out)
{
  out << "[GROUPS]" << endl;
  
  // Is there a way to obtain an iterator for a map that doesn't
  // break the const qualifier??
  map<string, string>::iterator i;
  
  for (i = groups.begin(); i != groups.end(); i++)
  {
    out << i->first << "=" << i->second << endl;
  }
  
  out << endl;
}

/**
 * Writes out the cut definition configuration to the supplied ostream.
 *
 * \param out The output stream.
 */
void Config::writeConfigTextCutDefinition(ostream& out) const
{
  for (unsigned int i = 0; i < cuts.size(); i++)
  {
    if (cuts[i].getExtension()->length() > 0)
    {
      out << "[" << cuts[i].getExtension()->substr(1) << "]" << endl;
      
      for (unsigned int j = 0; j < cuts[i].getStartREs()->size(); j++)
      {
        out << "START=/" << (*cuts[i].getStartREs())[j] << "/" << endl;
      }
      
      if (cuts[i].getStartREs()->size() == 0)
      {
        cerr << "ERROR! RE has no starting regular expression!" << endl;
      }
      
      for (unsigned int j = 0; j < cuts[i].getEndREs()->size(); j++)
      {
        out << "END=/" << (*cuts[i].getEndREs())[j] << "/" << endl;
      }
      
      if (cuts[i].getEndREs()->size() == 0)
      {
        out << "#END=" << endl;
      }
      
      out << "LENGTH=" << cuts[i].getMaxLength() << endl;
      
      if (cuts[i].getOutputDirectory()->length() > 0)
      {
        out << "DIRECTORY=" << (*cuts[i].getOutputDirectory()) << endl;
      } else
      {
        out << "#DIRECTORY=" << endl;
      }
      
      for (unsigned int j = 0; j < cuts[i].getComments()->size(); j++)
      {
        out << "COMMENT=" << (*cuts[i].getComments())[j] << endl;
      }
      
      if (cuts[i].getComments()->size() == 0)
      {
        out << "#COMMENT=" << endl;
      }
      
      out << endl;
    } else
    {
      cerr << "Error.  There is a RE in the config with no extension!" << endl;
    }
  }
}

/**
 * Converts a C++ string to upper case.  Perhaps not the best way of doing it.
 *
 * \param str The input string.
 * \returns The uppercase string.
 */
string Config::stringToUpper(string str) const
{
  for (unsigned int i = 0; i < str.length(); i++)
  {
    str[i] = toupper(str[i]);
  }
  
  return str;
}

/**
 * Activates the cut definition extensions supplied.
 *
 * Note that the supplied string may contain more than one extension.  If
 * so they will be separated by vertical pipes '|'.
 *
 * \param cut The string containing cut definition extensions.
 * \returns True if the operation is successful.  False otherwise.
 */
bool Config::activateCutDefinition(const string& cut)
{
  int allOk = true;
  int pos = 0;
  int lastPos = 0;
  
  do
  {
    pos = cut.find_first_of("|", lastPos);
    
    string filetype = cut.substr(lastPos, pos - lastPos);
    
    bool found = false;
    for (unsigned int i = 0; i < activeCuts.size(); i++)
    {
      if (strcasecmp(filetype.c_str(), activeCuts[i].getExtension()->substr(1).c_str()) == 0)
      {
        // The cut is already active.  No need to add it to the active list again.
        found = true;
        i = activeCuts.size();
      }
    }
    
    if (found == false)
    {
      for (unsigned int i = 0; i < cuts.size(); i++)
      {
        if (strcasecmp(filetype.c_str(), cuts[i].getExtension()->substr(1).c_str()) == 0)
        {
          activeCuts.push_back(cuts[i]);
          found = true;
          i = cuts.size();
        }
      }
      
      if (found == false)
      {
        cerr << "Could not find a definition for the " << filetype << " cut definition." << endl;
        allOk = false;
      }
    }
    
    lastPos = pos + 1;
  } while (pos != -1);
  
  return allOk;
}
