%module exgrep

%{
#include "forensic.h"
#include "main.h"

  /* Define some exception classes that could be thrown */
  class configException {
  public:
    configException(char *string) {
      error = string;
    }

    string error;
  };

  class StopIter{
  public:
    StopIter() {
    }
  };

 /**
   * The Exgrep class is used to search through an input stream looking for cuts.
   *
   * A cut is simply a series of bytes which has been identified by a particular
   * cut definition.  Typically this is used to search for files from a raw disk
   * without using the filesystem.
   *
   * Exgrep expects to be passed a configuration file when it is first constructed.
   * It is then possible to feed it file references or file descriptors for it to
   * start searching.
   */
  class PyExgrep
  {
  public:
    PyExgrep(char *filename){
      Config *a = new Config();
      string tmp = filename;
      if (tmp.size()>0 && parse(a, tmp) == false) {
	throw configException ("Unable to parse config file");
	this->exgrep_obj = NULL;
	return;
      };
      this->exgrep_obj = new Exgrep(*a);
    };

    ~PyExgrep() {
      delete this->exgrep_obj;
    };
    
    PyExgrep *__iter__() {
      return this;
    }

    int process(char * inputFilename){
      string tmp;
      tmp = inputFilename;
      if(!exgrep_obj) {
	throw("Unable to parse config file");
	return(0);
      };
      return (exgrep_obj->process(tmp));
    };
     
    bool next() {
      int result;
      result = exgrep_obj->hasNext();

      if(!result) {
	throw StopIter();
      };
      return (result);
    };
    
    const char *getCutName(){
      return(exgrep_obj->getNextCutDefinition()->getExtension()->c_str());
    };
    
    const char* getBuffer(){
      return (exgrep_obj->getNextBuffer());
    };
    unsigned int getBufferLength(){
      return (exgrep_obj->getNextBufferLength());
    };
    unsigned int getOffset(){
      return(exgrep_obj->getNextCutOffset());
    };

  private:
    Exgrep *exgrep_obj;
  };

%}

%typemap(python,in) std::string
{
  if (PyString_Check ($source))
  {
    $target = new std::string((char *)PyString_AsString($source));
  }
  else
  {
    PyErr_SetString (PyExc_TypeError, "not a string");
    return NULL;
  }
}

%typemap(python,out) std::string
{
  $target = PyString_FromString($source->c_str());
}

%except(python) {
  try {
    $function
      } catch(configException s) {
	PyErr_SetString(PyExc_IOError,s.error.c_str());
	return NULL;
      } catch (StopIter) {
	PyErr_SetString(PyExc_StopIteration,"Iteration Complete");
	return NULL;
      };
};

  class PyExgrep {
  public:
    PyExgrep(char *filename);
    ~PyExgrep();

    PyExgrep *__iter__();
    int process(char * inputFilename);
    bool next();
    const char *getCutName();
    const char* getBuffer();
    unsigned int getBufferLength();
    unsigned int getOffset();
  };
