/* 
Index tools are designed to index a dictionary of search terms in a
very large input file.

Algorithm:
-------------

Terminology: The text to be searched is referenced as 'text' The list
of words to locate within the text is called the 'dictionary'

This algorithm searches the text at approximately constant time with
respect to the number of words in the dictionary. (i.e. we can have an
(almost) arbitrarily large dictionary to index it in constant time).

The basic method for indexing occurances of words is by assigning 
each word a unique list of offsets into the text:
*/


/* heap_ptrs are essentially offsets into our heap where the data may
be found:

 before accessing the data as a struct ,you must use
 absolute_node(heap_ptr) to convert from the offsets into the heap to
 an actual pointer
*/
typedef unsigned int heap_ptr;

struct offset_list {
  unsigned long long int offset;
  heap_ptr next;
};

/*
Before indexing, we construct a tree in memory representing the
dictionary. Each node in this tree is a letter of a word in the
dictionary, and each node has an arbitrary number of child nodes
representing the letters following them:
*/
/* All the unsigned ints are references from the start of the heap */
struct idx_node {
  unsigned char c;
  heap_ptr offsets;
  heap_ptr child;
  heap_ptr peer;
};

/* 
If this node represents the end of the word, offset will be non-null
and refer to a valid offset list.

If there is a char following this char in a word, child will point to
the relevant child node.

Peer will point to this node's peer (i.e. the nodes which share the
same parent as this node). By following the peer pointer it is
possible to visit all peers.

Example:
suppost the dictionary holds the words ab,ac, acd

The first node corresponds to the letter a, and has a child node b. b
in turn has a peer c, but no children. Hence we can see that both b
and c are children to a. Both these nodes point at offset lists
storing their occurance in the text, but the c node also has another
child node d.

Indexing Algorithm:
-----------------------

Rather than search for the words in the text, we turn the tables
around and search for the text in the dictionary. Since the dictionary
is perfectly indexed this should be very fast:

1) initialise the algorithm with the root node, and the first char in the text.
2) Search for the char in the node's children
3) If not found, 

1) For each character in the text we search to see if its in the root node's children.
2) If it is, we read the next char from the file, and goto 1 with the next char and the 
*/
