# ******************************************************
# Copyright 2004: Commonwealth of Australia.
#
# Developed by the Computer Network Vulnerability Team,
# Information Security Group.
# Department of Defence.
#
# Michael Cohen <scudette@users.sourceforge.net>
# David Collett <daveco@users.sourceforge.net>
#
# ******************************************************
#  Version: FLAG 0.4 (12-02-2004)
# ******************************************************
#
# * This program is free software; you can redistribute it and/or
# * modify it under the terms of the GNU General Public License
# * as published by the Free Software Foundation; either version 2
# * of the License, or (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ******************************************************

""" Flag plugin to load various forms of data into the case databases """
import re,os
import pyflag.Reports as Reports
import pyflag.FlagFramework as FlagFramework
import pyflag.conf
config=pyflag.conf.ConfObject()
import pyflag.DB as DB
import pyflag.LogFile as LogFile

description = "Load Data"
order = 20

class LoadPresetLog(Reports.report):
    """ Loads a log file into the database using preset type """
    parameters = {"table":"alphanum", "datafile":"filename", "log_preset":"sqlsafe", "final":"alphanum"}
    name="Load Preset Log File"
    description="Load Data from log file into Database using Preset"
    order=30

    def display(self,query,result):
        result.heading("Uploaded log file into database")
        result.para("Successfully uploaded file %s into case %s in table %s" % (query['datafile'],query['case'],query['table']))
        return result
    
    def progress(self,query,result):
        result.heading("Currently uploading log file into database")
        dbh = self.DBO(query['case'])
        dbh.execute("select count(*) as count from %s_log", (query['table']))
        tmp = dbh.fetch()
        try:
            result.para("Uploaded %s rows. " % tmp['count'])
        except TypeError,e:
            print "exception %s tmp is %s" %(e,tmp)

    def form(self, query, result):
        try:
            result.case_selector()
            result.meta_selector(config.FLAGDB,'Select preset type','log_preset',onclick="this.form.submit();")
            result.textfield("Enter Name for table:","table")
            tbl_sel = self.ui()
            tbl_sel.meta_selector(config.FLAGDB,'Insert into existing table','%s_log'%query['log_preset'])
            result.row("OR:", tbl_sel)
            tmp = self.ui(default = result)
            tmp.filebox()
            result.row("Select file to load:",tmp)

            log = LogFile.get_loader(query['log_preset'])
            #print log

            # show preview
            dbh = self.DBO(query['case'])
            temp_table = dbh.get_temp()
            
            log.load(query['case'],query['datafile'],temp_table, rows=3)

            # retrieve and display the temp table
            dbh.execute("select * from %s limit 1",temp_table)
            columns =[]
            for d in dbh.cursor.description:
                columns.append(d[0])
            preview_tbl = self.ui()
            preview_tbl.table(columns=columns,names=columns,links=[], table=temp_table, case=query['case'],simple=True)
            result.row(preview_tbl, colspan=5, bgcolor='lightgray')
            
            result.checkbox('Click here when finished','final','ok')
            
        except KeyError:
           pass

    def analyse(self, query):
        """ Load the log file into the table """
        log = LogFile.get_loader(query['log_preset'])

        # decide on table name
        if query.has_key('new_table'):
            del query['table']
            query['table'] = query['new_table']

        print query['table']
        log.load(query['case'],query['datafile'],'%s_log'%query['table'])
	dbh = self.DBO(query['case'])
        dbh.execute("INSERT INTO meta set property='logtable', value='%s_log'"% query['table'])

    def reset(self, query):
        dbh = self.DBO(query['case'])
        dbh.execute("DROP TABLE %s_log"% query['table'])


##class LoadLogFile(Reports.report):
##    """ Loads a log file into the database. """
##    parameters = {"table":"alphanum", "datafile":"filename", "field":"unique", "final":"alphanum"}
##    name="Load Log File"
##    description="Load Data from log file into Database"
##    order=40

##    def display(self,query,result):
##        result.heading("Uploaded log file into database")
##        result.para("Successfully uploaded file %s into case %s in table %s" % (query['datafile'],query['case'],query['table']))

##        return result

##    def progress(self,query,result):
##        result.heading("Currently uploading log file into database")
##        dbh = self.DBO(query['case'])
##        dbh.execute("select count(*) as count from %s", (query['table']))
##        tmp = dbh.fetch()
##        try:
##            result.para("Uploaded %s rows. " % tmp['count'])
##        except TypeError,e:
##            print "exception %s tmp is %s" %(e,tmp)

##    def form(self,query,result):
##        try:
##            result.start_table()
##            result.ruler()
##            # Step 1 - Allow user to specify database and name the new table to be created.
##            tmp = self.ui()
##            tmp.heading("Step 1:")
##            result.row(tmp,"Select Case name and name the table")
##            result.ruler()
##            result.case_selector()
##            result.textfield("Enter Name for table:","table")
##            result.ruler()

##            #Check if the user has supplied a name for the table: (if they did not we throw an exception and return the form so far - the overall effect is that the form will grow as the user enters more data.
##            query['table']
            
##            tmp = self.ui()
##            tmp.heading("Step 2:")
##            result.row(tmp,"Select the log file to load")
##            result.ruler()
##            #Create a new ui based on our current one
##            tmp = self.ui(default = result)
##            tmp.filebox()
##            result.row("Enter name of file to load:",tmp)

##            result.ruler()
##            tmp = self.ui()
##            tmp.heading("Step 3:")
##            result.row(tmp,"Choose log class")
##            result.ruler()
##            result.selector("Log Class",'log_type',"select name,description from log_types",(),case=None)
            
##            result.ruler()
##            tmp = self.ui()
##            tmp.heading("Step 4:")
##            result.row(tmp,"")
##            result.ruler()

##            #If there is no delimiter method - we choose simple
##            try:
##                query['delmethod']
##            except KeyError:
##                query['delmethod'] = 'simple'

##            result.radio("Field Separator Type:","delmethod",('simple','advanced','preset'),onclick="this.form.submit();")

##            if query['delmethod'] == 'simple':
##                result.selector("Simple Field Separator:",'delimiter',"select delimiter,description from simple_delimiters",(),case=None)
##            elif query['delmethod'] == 'advanced':
##                result.textfield("Enter Regular Expression:","delimiter")
##            elif query['delmethod'] == 'preset':
##                result.selector("Select Preset File Filter:",'preset',"select name,name from log_preset where log_type=%s",(query['log_type']),case=None)

##            #show the first 3 lines from the file:
##            fd = open(query['datafile'],'r')
##            count = 3
##            sample = ''
##            for line in fd:
##                line = line.strip()
##                #Dont count lines starting with a comment or empty lines
##                if not line or line.startswith('#'): continue
                
##                sample+=line+"\n"
##                count -=1
##                if count <=0:
##                    break

##            result.row(sample,bgcolor='lightgray',colspan=5)

##            #Step 5:
##            result.ruler()
##            tmp = self.ui(result)
##            tmp.heading("Step 5:")
##            result.row(tmp,"Select a pre-filter to use on the data")
##            result.ruler()

##            #Make up a list of available filters from the pre-filter object
##            f = prefilter().filters
##            x = []
##            y = []
##            for i in f.keys():
##                x.append(i)
##                y.append(f[i][1])
                
##            result.const_selector("Prefilter to use on the data:",'prefilter',x,y,size=4,multiple="multiple")


##            #Now show the user what the pre-filters do on the data:
##            result.row("Text processed from file (after applying pre-filters)",colspan=5)
##            sample = self.transform(sample,query)
##            result.row(sample,bgcolor='lightgray',colspan=5)

##            result.end_table()
##            result.start_table()
##            #Step 6:
##            result.ruler()
##            tmp = self.ui(result)
##            tmp.heading("Step 6:")
##            result.row(tmp,"Assign name to each field",align='left')
##            result.ruler()
            
##            #Show the split up lines:
##            tmp_table = self.ui(result)
##            tmp_table.start_table(border='1')
            
##            if query['delmethod'] == 'simple':
##                result.row("Got a simple delmethod")
##                for line in sample.splitlines():
##                    try:
##                        line_array = line.split(query['delimiter'])
##                    except (KeyError, AttributeError):
##                        line_array = line.split()

##                    tmp_table.row(*line_array)

##            #Now add a line of choices based on field* values (we use field* rather than a field array since the order is important, whereas HTTP claims that order is not important for arrays.
##            field = []
##            for i in range(len(line_array)):
##                tmp_ui = self.ui(tmp_table)
##                #If there is no fieldxx arg, we set the default to be ignore
##                try:
##                    query["field%u"% i]
##                except KeyError:
##                    query["field%u"% i] = 'ignore'
##                tmp_ui.selector('',"field%u"% i,"select name,description from %s_fields"%query['log_type'],())
##                field.append(tmp_ui)

##            tmp_table.row(*field)
##            tmp_table.end_table()
##            result.row(tmp_table,bgcolor='lightgray',colspan=5)
##            result.hidden('field','ok')
##            result.end_table()
##            result.checkbox('Click here when ready to upload','final','ok')

##            return result

##        #If any of the required paramters are not there, we pass the form back to the user to complete.
##        except KeyError,e:
##            print "Got KeyError",e
##            return result

##    def transform(self,string,query):
##        """ Apply the transformations specified in query's prefilter array on the string.

##        @arg string: String data to operate on
##        @arg query: A query_type object which may contain some prefilter parameters
##        @return: A transformed string """
##        p = prefilter()
##        for i in query.getarray('prefilter'):
##            #Call the relevant methods on the prefilter object:
##            string = p.filters[i][0](p,string)

##        return string

##    def analyse(self,query):
##        #Find the list of fields that should be inserted:
##        fields = []  #Numerical list of indexes
##        schema = {}
##        for i in query.keys():
##            if i.startswith('field') and query[i] != 'ignore':
##                try:
##                    fields.append(int(i[len('field'):]))
##                    schema[query[i]] = ''
##                except ValueError:
##                    pass

##        #Pre-calculate the insert statement here
##        insert_str = "insert into "+query['table']+" values(Null"+',%r'*(len(fields))+")"

##        fdbh = self.DBO(None)
##        fdbh.execute("Select name,type,`key` from `%s_fields` where name not like \"ignore\"",(query['log_type']))

##        tmp = []
##        keys = []
##        for r in fdbh:
##            if schema.has_key(r['name']):
##                tmp.append(" ".join((r['name'],r['type'])))
##                if r['key'] == 1:
##                    keys.append(r['name'])

##        cdbh = self.DBO(query['case'])
###        cdbh.execute("Create table if not exists "+query['table']+ " (id int auto_increment,"+",".join(tmp)+",key(id))",())
##        cdbh.execute("Create table if not exists %s (id int auto_increment,%s,key(id))",(query['table'],",".join(tmp)))
            
##        tmp =''
##        fd = open(query['datafile'],'r')
##        #Try to insert some lines:
##        if query['delmethod'] == 'simple':
##            for line in fd:
##                line = line.strip()
##                #Dont count lines starting with a comment or empty lines
##                if not line or line.startswith('#'): continue
                
##                line = self.transform(line,query)
##                try:
##                    line_array = line.split(query['delimiter'])
##                except (KeyError, AttributeError):
##                    line_array = line.split()
                    
##                    selected_array = []
                    
##                    for i in fields:
##                        selected_array.append(line_array[i])

##                    cdbh.execute(insert_str,selected_array)

##            cdbh.execute("insert into meta set property=%r,value=%r",(query['log_type'],query['table']))
##        for i in keys:
##            cdbh.execute( "alter table %s add key(%s);",(query['table'],i))


import pyflag.Ethereal as Ethereal

class LoadTcpDump(Reports.report):
    """ Loads a TCPDump using the modified ethereal. """
    parameters = {"datafile":"filename"}
    name = "Load Pcap file"
    description = "Load a network capture file into the case Database "
    order=30
    
    def form(self,query,result):
        result.start_table()
        result.ruler()
        tmp = self.ui()
        tmp.heading("Step 1:")
        result.row(tmp,"Select Case name to load Pcap file into")
        result.case_selector()
        result.ruler()
        tmp = self.ui()
        tmp.heading("Step 2:")
        result.row(tmp,"Select PCAP file to load")
        tmp = self.ui(default = result)
        tmp.filebox()
        result.row("Enter name of file to load:",tmp)
        result.ruler()
        tmp = self.ui()
        tmp.heading("Step 3:")
        result.row(tmp,"")
##        result.checkbox('Perform basic analysis','preanalyse','ok',checked=1)

    def analyse(self,query):
        Ethereal.load_sql(query['case'],query['datafile'])

        #Now load the indexes on the data:
        dbh = self.DBO(query['case'])
        index = (
            ('tcp','key_id'),
            ('tcp','tcp_srcport'),
            ('tcp','tcp_dstport'),
            ('ip','key_id'),
            ('ip','ip_dst'),
            ('ip','ip_src'),
            ('frame','key_id'),
            ('eth','key_id'),
            ('icmp','key_id'),
            ('dns','key_id'),
            ('http','key_id'),
            ('http','http_request_uri(20)'),
            ('pop','key_id'),
            ('pop','pop_req_command(20)'),
            ('smtp','key_id'),
            ('smtp','smtp_req_command(20)'),
            ('udp','key_id'),
            ('udp','udp_srcport'),
            ('udp','udp_dstport')
            )
        
        for x,y in index:
            dbh.execute("alter table %s add index(%s)",(x,y))

    def display(self,query,result):
        result.heading("Uploaded PCAP file %r to case %r" % (query['datafile'],query['case']))
                       
    def progress(self,query,result):
        result.heading("Currently uploading Pcap file to case %s" % query['case'])
        dbh = self.DBO(query['case'])
        dbh.execute('select count(*) from frame',())
        a = dbh.cursor.fetchone()
        result.para("Added %s packets already"%a[0])

    def reset(self,query):
        dbh = self.DBO(query['case'])
        dbh2 =  self.DBO(query['case'])
        dbh.execute('select value from meta where property = "tcpdump_table" group by value',())
        while 1:
            rs = dbh.fetch()
            if not rs: break
            dbh2.execute("drop table %s ",rs['value'])

        dbh.execute("delete from meta where property = 'tcpdump_table'",())

import pyflag.IO as IO

class LoadIOSource(Reports.report):
    """ Initialises and caches an IO Subsystem datasource into the database for
    subsequent use by other reports (eg. LoadFS and exgrep) """
    parameters = {"iosource":"sqlsafe","subsys":"iosubsystem"}
    name = "Load IO Data Source"
    description = "Load a data source into flag using IO subsystem"
    order = 10

    def form(self,query,result):
        result.start_table()
        try:
            result.case_selector()
            result.ruler()
            subsystems=IO.subsystems.keys()
            subsystems.sort()
            subsystems.reverse()
            result.const_selector("Select IO Subsystem",'subsys',subsystems,subsystems)
            #this will cause a form to be placed into result.
            fd=IO.IOFactory(query,result)
            result.textfield("Identifier","iosource")
        except KeyError:
            pass
        except IOError, e:
            result.row("Error: %s" % e, bgcolor=config.HILIGHT)

    def analyse(self,query):
        # cache serialised io options in the case mata table
        fd=IO.IOFactory(query)
        dbh = self.DBO(query['case'])
        dbh.execute("insert into meta set property=%r,value=%r", ('iosource',query['iosource']))
        dbh.execute("insert into meta set property=%r,value=%r",(query['iosource'],fd.get_options()))

    def display(self,query,result):
        result.heading("Data Source %s successfully added" % query['iosource'])
        result.link("Load Filesystem", FlagFramework.query_type((), case=query['case'], family="LoadData", report="LoadFS", iosource=query['iosource']))
        result.para('')
        result.link("Extract Files", FlagFramework.query_type((), case=query['case'], family="UnstructuredDisk", report="ExtractFiles", iosource=query['iosource']))


import pyflag.Sleuthkit as Sleuthkit

class LoadFS(Reports.report):
    """ Loads Filesystem Image using the modified sleuthkit. """
    parameters = {"iosource":"iosource","fstype":"sqlsafe"}
    name = "Load Filesystem image"
    description = "Load a filesystem image into the case Database"
    order = 20
    
    def form(self,query,result):
        result.start_table()
        try:
            result.case_selector()
            result.ruler()
            result.meta_selector(message='Select IO Data Source', case=query['case'], property='iosource')
            
            # initialise/open the subsystem
            dbh = self.DBO(query['case'])
            fd=IO.open(query['case'],query['iosource'])
            
            magic = FlagFramework.Magic()
            result.ruler()
            result.row("Magic identifies this file as: %s" % magic.buffer(fd.read(10240)),colspan=50,bgcolor=config.HILIGHT)
            fd.close()
            result.const_selector("Enter Filesystem type",'fstype',['linux-ext2','linux-ext3','bsdi','fat','fat12','fat16','fat32','freebsd','netbsd','ntfs','openbsd','solaris'],['Linux ext2','Linux ext3','BSDi FFS','auto-detect FAT','FAT12','FAT16','FAT32','FreeBSD FFS','NetBSD FFS','NTFS','OpenBSD FFS','Solaris FFS'])
            result.ruler()
#            result.textfield("Identifier (eg. mount point)","identifier")
        except (KeyError,IOError):
            pass

    def analyse(self,query):
        """ load the filesystem image data into the database """
        dbh = self.DBO(query['case'])
        tablename=dbh.MakeSQLSafe(query['iosource'])
        # call on sleuthkit module to shell out and load data
        Sleuthkit.load_sleuth(query['case'],query['fstype'],tablename,query['iosource'])
        dbh.execute("insert into meta set property='fsimage', value=%r", query['iosource'])
#        dbh.execute("insert into meta set property=%r, value=%r", (tablename, query['iosource']))
        
        #Add indexes:
        index = (
            ('file','inode'),
            ('file','path(100)'),
            ('file','name(100)'),
            ('inode','inode'),
            ('block','inode')
            )
        for x,y in index:
            dbh.execute("alter table %s_%s add index(%s)",(x,tablename,y))
        
    def display(self,query,result):
        result.heading("Uploaded FS Image from IO Source %s to case %s" % (query['iosource'],query['case']))
        result.link("Analyse this data", FlagFramework.query_type((), case=query['case'], family="DiskForensics", fsimage=query['iosource']))
                       
    def progress(self,query,result):
        result.heading("Uploading filesystem image to case %s" % query['case'])
        dbh = self.DBO(query['case'])
        tablename=dbh.MakeSQLSafe(query['iosource'])
        try:
            result.start_table()
            dbh.execute("select count(*) from file_%s" % tablename)
            f = dbh.cursor.fetchone()
            result.row("Uploaded File Entries:", "%s"%f[0])
            dbh.execute("select count(*) as count,value as total from inode_%s, meta_%s as m where m.name='last_inode' group by total" % (tablename, tablename))
            row = dbh.fetch()
            result.row("Uploaded Inode Entries:", "%s of %s"%(row['count'],row['total']))
            result.end_table()
        except (TypeError, DB.DBError):
            pass

    def reset(self,query):
    	dbh = self.DBO(query['case'])
        tablename = dbh.MakeSQLSafe(query['identifier'])
        Sleuthkit.del_sleuth(query['case'],tablename)

class LoadKB(LoadTcpDump):
    """ Calculates a Knowledgebase based on a Pcap file and stores it in the case """
    parameters = {"datafile":"filename"}
    name = "Build Knowledge Base"
    description = "Builds a knowledge base from a Pcap file "
    order=50
    
    def form(self,query,result):
        result.start_table()
        result.ruler()
        tmp = self.ui()
        tmp.heading("Step 1:")
        result.row(tmp,"Select Case name to load Pcap file into")
        result.case_selector()
        result.ruler()
        tmp = self.ui()
        tmp.heading("Step 2:")
        result.row(tmp,"Select PCAP file to load")
        tmp = self.ui(default = result)
        tmp.filebox()
        result.row("Enter name of file to load:",tmp)
        
    def analyse(self,query):
        Ethereal.load_kb(query['case'],query['datafile'])

    def display(self,query,result):
        result.heading("Calculated the Knowledge Base in case %s " % query['case'])
        link = self.ui(result)
        link.link('Analyse this data',FlagFramework.query_type((),family='KnowledgeBase'))
        result.para(link)

    def progress(self,query,result):
        result.heading("Currently Building Knowledge Base into case %s" % query['case'])
        dbh = self.DBO(query['case'])
        dbh.execute('select count(*) from knowledge_node',())
        a = dbh.cursor.fetchone()
        result.para("Added %s nodes already"%a[0])

    def reset(self,query):
        Ethereal.clear_kb(query['case'])

