""" An extracting Grep implementation

This module will extract files from an image by using their magic.
"""
import re
import pyflag.conf
config=pyflag.conf.ConfObject()

## This initialises the cut definition stack:
definitions=[]

cut={}
cut["Extension"]="jpg"
cut["StartRE"]="\\xff\\xd8....(JFIF|Exif)"
cut["MaxLength"]=500000
cut["Comment"]="JPEG picture file type"
definitions.append(cut)

cut={}
cut["Extension"]="gif"
cut["StartRE"]="GIF8[79]a"
cut["MaxLength"]=50000
cut["Comment"]="GIF picture file type"
definitions.append(cut)

cut={}
cut["Extension"]="png"
cut["StartRE"]="\\x89PNG\\x0d\\x0a\\x1a\\x0a"
cut["EndRE"]="\\x45\\x4e\\x44\\xae\\x42\\x60\\x82"
cut["MaxLength"]=500000
cut["Comment"]="PNG picture file type"
definitions.append(cut)

cut={}
cut["Extension"]="tif"
cut["StartRE"]="\\x49\\x49\\x2a\\x00\\x08"
cut["MaxLength"]=1000000
cut["Comment"]="TIF picture file type 1"
definitions.append(cut)

cut={}
cut["Extension"]="tif"
cut["StartRE"]="\\x4d\\x4d\\x00\\x2a\\x00"
cut["MaxLength"]=1000000
cut["Comment"]="TIF picture file type 2"
definitions.append(cut)

cut={}
cut["Extension"]="doc"
cut["StartRE"]="\\xd0\\xcf\\x11\\xe0"
cut["MaxLength"]=500000
cut["Comment"]="MS Word document"
definitions.append(cut)

cut={}
cut["Extension"]="pdf"
cut["StartRE"]="%PDF-"
cut["EndRE"]=".%%EOF\\x0d"
cut["MaxLength"]=1000000
cut["Comment"]="Portable Document Format"
definitions.append(cut)

cut={}
cut["Extension"]="eps"
cut["StartRE"]="%!PS-Adobe"
cut["EndRE"]="end.%%.trailer"
cut["MaxLength"]=1000000
cut['Comment']='Encapsulated Postscript'
definitions.append(cut)

cut={}
cut["Extension"]="eps"
cut["StartRE"]="%!PS-Adobe"
cut["EndRE"]="%%EOF."
cut["MaxLength"]=1000000
cut['Comment']='Encapsulated Postscript'
definitions.append(cut)

cut={}
cut["Extension"]="ie_hist"
cut["StartRE"]="Client UrlCache"
cut["MaxLength"]=300000
cut["Comment"]="Internet Explorer URL cache"
definitions.append(cut)

cut={}
cut["Extension"]="url"
cut["StartRE"]="URL \\x03\\x00\\x00\\x00"
cut["MaxLength"]=384
cut["Comment"]="Internet Explorer URL cache"
definitions.append(cut)

cut={}
cut["Extension"]="url"
cut["StartRE"]="URL \\x02\\x00\\x00\\x00"
cut["MaxLength"]=256
cut["Comment"]="Internet Explorer URL cache"
definitions.append(cut)

cut={}
cut["Extension"]="mov"
cut["StartRE"]="....free.....mov"
cut["MaxLength"]=1000000
cut["Comment"]="Quicktime MOV format type 1"
definitions.append(cut)

cut={}
cut["Extension"]="mov"
cut["StartRE"]="MOVI"
cut["MaxLength"]=1000000
cut["Comment"]="Quicktime MOV format type 2"
definitions.append(cut)

cut={}
cut["Extension"]="mov"
cut["StartRE"]="....moov"
cut["MaxLength"]=1000000
cut["Comment"]="Quicktime MOV format type 3"
definitions.append(cut)

cut={}
cut["Extension"]="mov"
cut["StartRE"]="....mdat"
cut["MaxLength"]=1000000
cut["Comment"]="Quicktime MOV format type 4"
definitions.append(cut)

cut={}
cut["Extension"]="avi"
cut["StartRE"]="RIFF....AVI"
cut["MaxLength"]=1000000
cut["Comment"]="AVI video format"
definitions.append(cut)

cut={}
cut["Extension"]="wmv"
cut["StartRE"]="\\x30\\x26\\xb2\\x75\\x8e\\x66"
cut["MaxLength"]=1000000
cut["Comment"]="Windows movie file"
definitions.append(cut)

for i in definitions:
    i["CStartRE"]=re.compile(i["StartRE"])
    try:
        i["CEndRE"]=re.compile(i["EndRE"])
    except: pass

import pyflag.IO as IO

def process(case,subsys):
    """ A generator to produce all the recoverable files within the io object identified by identifier """
    io=IO.open(case,subsys)
    blocksize=1024*1024
    windowsize=100
    count=0
    window=''
    while(1):
        ## This implements a sliding window of window bytes to ensure
        ## we do not miss a signature that was split across blocksize:
        try:
            data=io.read(blocksize)
            if not data: break
        except IOError:
            break
        
        f=window+data
        for cut in definitions:
            pos=0
            while pos<blocksize:
                match=cut['CStartRE'].search(f,pos)
                if match:
                    yield({'offset':match.start()+count-len(window),'length':cut['MaxLength'],'type':cut['Extension']})
                    pos=match.start()+1
                else:
                    pos=blocksize

        window=f[-windowsize:]
        count+=blocksize
        
    io.close()
