# ******************************************************
# Copyright 2004: Commonwealth of Australia.
#
# Developed by the Computer Network Vulnerability Team,
# Information Security Group.
# Department of Defence.
#
# Michael Cohen <scudette@users.sourceforge.net>
#
# ******************************************************
#  Version: FLAG 0.4 (12-02-2004)
# ******************************************************
#
# * This program is free software; you can redistribute it and/or
# * modify it under the terms of the GNU General Public License
# * as published by the Free Software Foundation; either version 2
# * of the License, or (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ******************************************************

""" Main UI module.

The output within flag is abstracted such that it is possible to connect any GUI backend with any GUI Front end. This is done by use of UI objects. When a report runs, it will generate a UI object, which will be built during report execution. The report then returns the object to the calling framework which will know how to handle it. Therefore the report doesnt really know or care how the GUI is constructed """

import re,cgi,types
import pyflag.FlagFramework as FlagFramework
import pyflag.DB as DB
import pyflag.conf
import pyflag.UI as UI
config=pyflag.conf.ConfObject()

class GTKUI(UI.GenericUI):
    """ A GTK UI Implementation. """
    main_widget = None
    widgets = []
    flag = None

    def __init__(self):
        import gtk
        self.widgets = []
        return

    def heading(self,string):
        label = gtk.Label(string)
        self.widgets.append(label)

    def __str__(self):
        return "GTK Widget"
    
    def display(self):
        container = gtk.VBox()
        for widget in self.widgets:
            container.add(widget)

        return container

    tables = []
    def start_table(self,**options):
        self.tables.append([])

    def row(self,*columns, **options):
        if not self.tables:
            self.start_table()

        hbox = gtk.HBox()
        for column in columns:
            if isinstance(column,str):
                column = gtk.Label(column)
            elif isinstance(column,GTKUI):
                column = column.display()
                
            hbox.add(column)

        self.widgets.append(hbox)

        def end_table(self):
            pass

    def goto_link(self,widget):
        if self.link_callback:
            self.link_callback(widget.get_data('query'))

    def link(self,string,target=FlagFramework.query_type(()),**target_options):
        import pyflag.FlagFramework as FlagFramework

        if target_options:
            for k,v in target_options.items():
                k = FlagFramework.urlencode(k)
                v = FlagFramework.urlencode(v)
                target += "%s=%s&" % (k,v)


#        label = gtk.Label(string)
#        label.set_data('query',target)
#        label.connect("button-press-event",self.goto_link)
#        self.widgets.append(label)
        button = gtk.Button(string)
        button.set_data('query',target)
        button.connect("clicked",self.goto_link)
        self.widgets.append(button)
